from __future__ import print_function
import argparse
from getpass import getpass
import logging
import os

import colorama
import pyperclip
from tabulate import tabulate

from .db import Database
from .credential import Credential, CredentialNotFoundError

try:
    input = raw_input
except NameError:
    pass

DEFAULT_DATABASE_PATH = os.path.join(
    os.path.expanduser("~"),
    ".pysswords"
)
DEFAULT_GPG_BINARY = "gpg"

colorama.init(autoreset=True)


def get_args(command_args=None):
    """Return args from command line"""
    parser = argparse.ArgumentParser(prog="pysswords")
    parser.add_argument("-I", "--init", action="store_true",
                        help="create a new Pysswords database")
    parser.add_argument("-d", "--database", default=DEFAULT_DATABASE_PATH,
                        metavar="<DATABASE PATH>",
                        help="specify path to database")
    parser.add_argument("--show-password", action="store_true",
                        help="show password as plain text when print")
    parser.add_argument("-a", "--add", action="store_true",
                        help="add new credential")
    parser.add_argument("-r", "--remove", metavar="<CREDENTIAL NAME>",
                        help="delete credential")
    parser.add_argument("-e", "--edit", metavar="<CREDENTIAL NAME>",
                        help="edit credential")
    parser.add_argument("-s", "--search", metavar="<QUERY>",
                        help="search credential")
    parser.add_argument("-g", "--get", metavar="<CREDENTIAL NAME>",
                        help="print credential")
    parser.add_argument("-l", "--list", action="store_true",
                        help="print all credentials as a table")
    parser.add_argument("-c", "--clipboard", action="store_true",
                        help="copy credential password to clipboard")
    parser.add_argument("--gpg", metavar="<GPG>", default=DEFAULT_GPG_BINARY,
                        help="gpg binary name")
    args = parser.parse_args(command_args)

    if args.clipboard and not args.get:
        parser.error('-g argument is required in when using -c')
    return args


def get_password(prompt="Password: "):
    for _ in range(3):
        password = getpass(prompt)
        repeat_password = getpass("Type again: ")

        if password == repeat_password:
            return password
        else:
            print("Entries don't match!")
    else:
        raise ValueError("Entries didn't match")


def check_passphrase(database, passphrase):
    sig = database.gpg.sign(
        'testing',
        default_key=database.gpg_key,
        passphrase=passphrase
    )
    if not sig:
        msg = "Wrong passphrase for database at '{}'".format(database.path)
        raise ValueError(msg)
    return sig


def list_credentials(database, query=None, show_password=False):
    if show_password:
        passphrase = getpass("Database passphrase: ")
        check_passphrase(database, passphrase)
    headers = ["name", "login", "password", "comments"]
    table = []
    for credential in database.credentials:
        row = [
            colorama.Fore.YELLOW + credential.name + colorama.Fore.RESET,
            credential.login,
            "..." if not show_password else database.gpg.decrypt(
                credential.password,
                passphrase=passphrase
            ),
            credential.comments
        ]
        table.append(row)
    print(tabulate(table, headers, tablefmt="orgtbl"))


def prompt_credential(**defaults):
    credential_name = input("Name: ")
    credential_login = input("Login: ")
    credential_password = get_password("Credential password: ")
    credential_comments = input("Comments [optional]: ")
    return Credential(
        name=credential_name,
        login=credential_login,
        password=credential_password,
        comments=credential_comments,
    )


def add_credential(database):
    credential = prompt_credential()
    database.add(credential)


def copy_password_to_clipboard(database, credential_name):
    credential = database.credential(name=credential_name)
    if credential:
        passphrase = getpass("Database passphrase: ")
        check_passphrase(database, passphrase)
        pyperclip.copy(
            database.gpg.decrypt(credential.password, passphrase=passphrase)
        )
        print("Password for '{}' copied to clipboard".format(credential.name))


def get_confirmation(prompt):
    prompt = "{} (y|n): ".format(prompt)
    return input(prompt)


def remove_credential(database, name):
    credential = database.credential(name=name)
    prompt = "Remove `{}`".format(credential)
    answer = get_confirmation(prompt)
    if answer and answer.lower()[0] == "y":
        database.remove(name=name)


def get_credential(database, name, to_clipboard=False):
    try:
        credential = database.credential(name=name)
        if to_clipboard:
            copy_password_to_clipboard(database=database,
                                       credential_name=name)
        else:
            print(credential)
    except CredentialNotFoundError:
        logging.info("Credential was not found")


def edit_credential(database, name):
    credential = database.credential(name=name)
    prompt = "Edit `{}`".format(credential)
    answer = get_confirmation(prompt)
    if answer and answer.lower()[0] == "y":
        edited_credential = prompt_credential()
        values = {
            "name": edited_credential.name,
            "login": edited_credential.login,
            "password": edited_credential.password,
            "comments": edited_credential.comments,
        }
        database.edit(name=name, values=values)


def run(args=None):
    args = get_args() if args is None else args

    if args.init:
        database = Database.create(
            path=args.database,
            passphrase=get_password("Database passphrase: "),
            gpg_bin=args.gpg
        )
        logging.info("Database created at '{}'".format(database.path))
    else:
        database = Database.from_path(path=args.database,
                                      gpg_bin=args.gpg)
        if args.add:
            add_credential(database)
        elif args.get:
            get_credential(database,
                           name=args.get,
                           to_clipboard=args.clipboard)
        elif args.remove:
            remove_credential(database, name=args.remove)
        elif args.edit:
            edit_credential(database, name=args.edit)
        elif args.search:
            list_credentials(database=database,
                             query=args.search,
                             show_password=args.show_password)
        else:
            list_credentials(database=database,
                             show_password=args.show_password)


if __name__ == "__main__":
    run()



import os


def touch(fname):
    """Mimics behavior of Unix touch command"""
    open(fname, "a").close()


def which(program):
    """Mimics behavior of UNIX which command. """
    # Add .exe program extension for windows support
    if os.name == "nt" and not program.endswith(".exe"):
        program += ".exe"

    for envdir in os.environ["PATH"].split(os.pathsep):
        program_path = os.path.join(envdir, program)
        if os.path.isfile(program_path) and os.access(program_path, os.X_OK):
            return program_path



import os


class CredentialNotFoundError(ValueError):
    pass


class Credential(object):
    def __init__(self, name, login, password, comments):
        self.name = name
        self.login = login
        self.password = password
        self.comments = comments

    def save(self, database_path):
        credential_path = os.path.join(database_path, self.name)
        os.makedirs(credential_path)

        with open(os.path.join(credential_path, "login"), "w") as f:
            f.write(self.login)

        with open(os.path.join(credential_path, "password"), "w") as f:
            f.write(self.password)

        with open(os.path.join(credential_path, "comments"), "w") as f:
            f.write(self.comments)

    @classmethod
    def from_path(cls, path):
        return Credential(
            name=os.path.basename(path),
            login=open(os.path.join(path, "login")).read(),
            password=open(os.path.join(path, "password")).read(),
            comments=open(os.path.join(path, "comments")).read()
        )

    def __str__(self):
        return "Credential(name={}, login={})".format(
            self.name,
            self.login
        )



from glob import glob
import os
import shutil

from .credential import Credential, CredentialNotFoundError
from .crypt import create_gpg, load_gpg

try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError


class Database(object):

    def __init__(self, path, gpg):
        self.path = path
        self.gpg = gpg

    @classmethod
    def create(cls, path, passphrase, gpg_bin="gpg"):
        gpg = create_gpg(gpg_bin, path, passphrase)
        return Database(path, gpg)

    @classmethod
    def from_path(cls, path, gpg_bin="gpg"):
        gpg = load_gpg(binary=gpg_bin, database_path=path)
        return Database(path, gpg)

    @property
    def gpg_key(self):
        return self.gpg.list_keys(secret=True)[0]["fingerprint"]

    @property
    def credentials(self):
        return [self.credential(os.path.basename(c))
                for c in glob(self.path + "/**")]

    def add(self, credential):
        encrypted_password = self.gpg.encrypt(
            credential.password,
            self.gpg_key,
            cipher_algo="AES256"
        )
        credential.password = str(encrypted_password)
        credential.save(database_path=self.path)

    def remove(self, name):
        credential_path = os.path.join(self.path, name)
        shutil.rmtree(credential_path)

    def edit(self, name, values):
        credential = self.credential(name)
        new_credential = Credential(
            name=values.get("name", credential.name),
            login=values.get("login", credential.login),
            password=values.get("password", credential.password),
            comments=values.get("comments", credential.comments),
        )
        self.remove(name=name)
        self.add(new_credential)

    def credential(self, name):
        credential_path = os.path.join(self.path, name)
        try:
            credential = Credential.from_path(credential_path)
        except FileNotFoundError:
            raise CredentialNotFoundError(
                "Credential was not found in the database")
        return credential

    def search(self, query):
        return [c for c in self.credentials
                if (query in c.name) or
                (query in c.login) or
                (query in c.comments)]



import os
import gnupg

from .utils import which


def create_key_input(gpg, passphrase, testing=False):
    key_input = gpg.gen_key_input(
        name_real='Pysswords',
        name_email='pysswords@pysswords',
        name_comment='Autogenerated by Pysswords',
        passphrase=passphrase,
        testing=testing
    )
    return key_input


def create_gpg(binary, database_path, passphrase):
    gnupg_path = os.path.join(database_path, ".gnupg")
    gpg = gnupg.GPG(which(binary), homedir=gnupg_path)
    gpg.gen_key(create_key_input(gpg, passphrase))
    return gpg


def load_gpg(binary, database_path):
    gnupg_path = os.path.join(database_path, ".gnupg")
    gpg = gnupg.GPG(which(binary), homedir=gnupg_path)
    return gpg
